using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows;
using Microsoft.Build.BuildEngine;

namespace CheckSLStyles
{
    /// <summary>
    /// Analyse Stlye usage in a a Silverlight solution
    /// </summary>
    internal class SolutionAnalyser
    {
        private Solution _solution;
        private string _filename;
        private List<string> _csFiles = new List<string>();
        private List<string> _xamlFiles = new List<string>();
        private List<string> _warnings = new List<string>();
        private List<KeyValuePair<string,string>> _styles = new List<KeyValuePair<string,string>>();

        public int StylesFound { get; private set; }

        public SolutionAnalyser(string filename)
        {
            _filename = filename;
            _solution = new Solution(filename);
        }

        /// <summary>
        /// Analyse a solution
        /// </summary>
        /// <returns>Return the analysis</returns>
        public string Analyse()
        {
            foreach (var project in _solution.Projects)
                ReadProjectFiles(Path.Combine(Path.GetDirectoryName(_filename), project));
            ReadProjectStyles();
            CheckStylesUsage();
            return OutputNonUsedStyles();
        }

        /// <summary>
        /// Read all styles from all Silverligh xaml files
        /// </summary>
        private void ReadProjectStyles()
        {
            foreach (var xaml in _xamlFiles)
            {
                var lines = File.ReadAllText(xaml);

                // Look for style definition based on Regex match
                var matches = Regex.Matches(lines, "<(Style|[\\w]*Brush).*x:Key=\"(?<styleName>[\\w]*?)\".*>");
                if (matches.Count > 0)
                {
                    foreach (Match match in matches)
                    {
                        Group grp = match.Groups["styleName"];
                        if (grp != null && !string.IsNullOrEmpty(grp.Value))
                        {
                            // Add style to our collection
                            _styles.Add(new KeyValuePair<string, string>(grp.Value, xaml));
                        }
                    }
                }
            }
            StylesFound = _styles.Count;
        }

        /// <summary>
        /// Output to a string the non used styles
        /// </summary>
        /// <returns></returns>
        private string OutputNonUsedStyles()
        {
            var sb = new StringBuilder();
            if (_warnings.Any())
            {
                sb.AppendLine("Warnings:");
                foreach (var w in _warnings)
                    sb.AppendLine(string.Format("  - {0}", w));
                sb.AppendLine();
            }

            sb.AppendLine(string.Format("{0} styles found", StylesFound));
            sb.AppendLine(string.Format("Non-referenced styles {0}:", _styles.Count));

            foreach (var s in _styles)
                sb.AppendLine(string.Format("{0} \t\t\ton\t\t\t{1}", s.Key, s.Value));

            try
            {
                Clipboard.SetData(DataFormats.Text, sb.ToString());
            }
            catch (Exception)
            {
                sb.Insert(0, "Could not copy these results to the Clipboard!!!\r\n\r\n");
            }

            return sb.ToString();
        }

        /// <summary>
        /// Verify which styles are being used
        /// </summary>
        private void CheckStylesUsage()
        {
            foreach (var xaml in _xamlFiles)
            {
                var lines = File.ReadAllText(xaml);

                // Look for style usage based on a Regex match
                var matches = Regex.Matches(lines, "Resource (ResourceKey=\")?(?<styleName>[\\w]*?)(\\}|\")");
                if (matches.Count > 0)
                {
                    foreach (Match match in matches)
                    {
                        Group grp = match.Groups["styleName"];
                        if (grp != null && !string.IsNullOrEmpty(grp.Value))
                        {
                            int removed = _styles.RemoveAll(n => n.Key == grp.Value);
                        }  
                    }
                }
            }
        }

        /// <summary>
        /// read a project file (.csproj) and extract the relevant files
        /// </summary>
        /// <param name="projectFilename">File to read</param>
        private void ReadProjectFiles(string projectFilename)
        {
            Project project = new Project();

            if (!File.Exists(projectFilename))
            {
                _warnings.Add(string.Format("File not found {0}", projectFilename));
                return;
            }

            try
            {
                project.Load(projectFilename);
                var property = project.EvaluatedProperties["SilverlightApplication"];

                bool isSl = property != null && property.Value == "true";

                if (!isSl)
                {
                    isSl = project.Imports.Cast<Import>().Any(n => n.ProjectPath.Contains("Microsoft.Silverlight.CSharp.targets"));
                    if (!isSl)
                        return;
                }

                string path = Path.GetDirectoryName(projectFilename) ?? "";

                // Read Xaml files
                foreach (var grp in project.ItemGroups.Cast<BuildItemGroup>())
                    foreach (var item in grp.Cast<BuildItem>())
                        if (item.Name == "Compile")
                        {
                            if (!string.IsNullOrEmpty(item.Include) && item.Include.EndsWith(".cs", StringComparison.InvariantCultureIgnoreCase))
                            {
                                string csFile = Path.Combine(path, item.Include);
                                _csFiles.Add(csFile);
                                var xaml = item.GetEvaluatedMetadata("DependentUpon");
                                if (!string.IsNullOrEmpty(xaml) && xaml.EndsWith(".xaml", StringComparison.InvariantCultureIgnoreCase))
                                    _xamlFiles.Add(Path.Combine(Path.GetDirectoryName(csFile) ?? "", xaml));
                            }
                        }
                        else if (item.Name == "Page")
                        {
                            if (!string.IsNullOrEmpty(item.Include) && item.Include.EndsWith(".xaml", StringComparison.InvariantCultureIgnoreCase))
                            {
                                string xaml = Path.Combine(path, item.Include);
                                _xamlFiles.Add(xaml);
                            }
                        }
            }
            catch (Exception e)
            {
                _warnings.Add(string.Format("Error parsing project {0}: {1}", projectFilename, e.Message));
            }
        }
    }
}